#!/usr/bin/env python

# Copyright (C) 2019 Vanessa Sochat.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sif
import argparse
import sys
import os


def get_parser():
    parser = argparse.ArgumentParser(description="SIF Python")

    # Global Variables
    parser.add_argument('--debug', dest="debug", 
                        help="use verbose logging to debug.", 
                        default=False, action='store_true')

    parser.add_argument('--quiet', dest="quiet", 
                        help="show SIF Python verison and exit", 
                        default=False, action='store_true')

    parser.add_argument('--version', dest="version", 
                        help="suppress additional output.", 
                        default=False, action='store_true')

    description = 'actions for SIF Python'
    subparsers = parser.add_subparsers(help='sif python actions',
                                       title='actions',
                                       description=description,
                                       dest="command")


    # Local shell with client loaded
    shell = subparsers.add_parser("shell",
                                  help="shell into a session a client.")

    shell.add_argument("image", nargs=1,
                       help="the image to load into the client", 
                       type=str, default=None)

    return parser


def get_subparsers(parser):
    '''get_subparser will get a dictionary of subparsers, to help with printing help
    '''

    actions = [action for action in parser._actions 
               if isinstance(action, argparse._SubParsersAction)]

    subparsers = dict()
    for action in actions:
        # get all subparsers and print help
        for choice, subparser in action.choices.items():
            subparsers[choice] = subparser

    return subparsers



def main():
    '''main is the entrypoint to the sregistry client. The flow works to first
    to determine the subparser in use based on the command. The command then
    imports the correct main (files imported in this folder) associated with
    the action of choice. When the client is imported, it is actually importing
    a return of the function get_client() under sregistry/main, which plays
    the job of "sniffing" the environment to determine what flavor of client
    the user wants to activate. Installed within a singularity image, this
    start up style maps well to Standard Container Integration Format (SCIF)
    apps, where each client is a different entrypoint activated based on the
    environment variables.
    '''

    parser = get_parser()
    subparsers = get_subparsers(parser)

    def help(return_code=0):
        '''print help, including the software version and active client 
           and exit with return code.
        '''

        version = sif.__version__

        print("\nSIF Python v%s" % version)
        parser.print_help()
        sys.exit(return_code)
    
    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()
    try:
        args = parser.parse_args()
    except:
        sys.exit(0)

    if args.debug is False:
        os.environ['MESSAGELEVEL'] = "DEBUG"

    # Show the version and exit
    if args.version is True:
        print(sif.__version__)
        sys.exit(0)

    from sif.logger import bot

    # Does the user want a shell?
    if args.command == "shell": from .shell import main

    # Pass on to the correct parser
    return_code = 0
    try:
        main(args)
        sys.exit(return_code)
    except UnboundLocalError:
        return_code = 1

    help(return_code)

if __name__ == '__main__':
    main()



'''

Copyright (C) 2018-2019 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

import datetime
import hashlib
import errno
import os
import pwd
import re
import shutil
import tempfile
import tarfile

import json
from sif.utils import get_installdir
from sif.logger import bot
import io
import sys


################################################################################
## FOLDER OPERATIONS ###########################################################
################################################################################


def mkdir_p(path):
    '''mkdir_p attempts to get the same functionality as mkdir -p

       Parameters
       =========
       path: the path to create.
    '''
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            bot.error("Error creating path %s, exiting." % path)
            sys.exit(1)


def get_tmpfile(requested_tmpdir=None, prefix=""):
    '''get a temporary file with an optional prefix. By default will be
       created in /tmp unless SIF_TMPDIR is set. By default, the file
       is closed (and just a name returned).

       Parameters
       ==========
       requested_tmpdir: an optional requested temporary directory, first
       priority as is coming from calling function.
       prefix: Given a need for a sandbox (or similar), prefix the file
       with this string.
    '''

    # First priority for the base goes to the user requested.
    tmpdir = get_tmpdir(requested_tmpdir)

    # If tmpdir is set, add to prefix
    if tmpdir is not None:
        prefix = os.path.join(tmpdir, os.path.basename(prefix))

    fd, tmp_file = tempfile.mkstemp(prefix=prefix) 
    os.close(fd)

    return tmp_file


def get_tmpdir(requested_tmpdir=None, prefix="", create=True):
    '''get a temporary directory for an operation. If SIF_TMPDIR
       is set, return that. Otherwise, return the output of tempfile.mkdtemp

       Parameters
       ==========
       requested_tmpdir: an optional requested temporary directory, first
       priority as is coming from calling function.
       prefix: Given a need for a sandbox (or similar), we will need to 
       create a subfolder *within* the SIF_TMPDIR.
       create: boolean to determine if we should create folder (True)
    '''
    from sif.defaults import SIF_TMPDIR

    # First priority for the base goes to the user requested.
    tmpdir = requested_tmpdir or SIF_TMPDIR

    prefix = prefix or "sif-tmp"
    prefix = "%s.%s" %(prefix, next(tempfile._get_candidate_names()))
    tmpdir = os.path.join(tmpdir, prefix)

    if not os.path.exists(tmpdir) and create is True:
        os.mkdir(tmpdir)

    return tmpdir

def get_userhome():
    '''get the user home based on the effective uid
    '''
    return pwd.getpwuid(os.getuid())[5]


################################################################################
## COMPRESSION #################################################################
################################################################################

def extract_tar(archive, output_folder, handle_whiteout=False):
    '''extract a tar archive to a specified output folder

        Parameters
        ==========
        archive: the archive file to extract
        output_folder: the output folder to extract to
        handle_whiteout: use docker2oci variation to handle whiteout files

    '''
    from .terminal import run_command

    # Do we want to remove whiteout files?
    if handle_whiteout is True:
        return _extract_tar(archive, output_folder)

    # If extension is .tar.gz, use -xzf
    args = '-xf'
    if archive.endswith(".tar.gz"):
        args = '-xzf'

    # Just use command line, more succinct.
    command = ["tar", args, archive, "-C", output_folder, "--exclude=dev/*"]
    if not bot.is_quiet():
        print("Extracting %s" % archive)

    return run_command(command)



################################################################################
## FILE OPERATIONS #############################################################
################################################################################

def write_file(filename, content, mode="w"):
    '''write_file will open a file, "filename" and write content, "content"
       and properly close the file
    '''
    with open(filename, mode) as filey:
        filey.writelines(content)
    return filename


def write_json(json_obj, filename, mode="w", print_pretty=True):
    '''write_json will (optionally,pretty print) a json object to file

       Parameters
       ==========
       json_obj: the dict to print to json
       filename: the output file to write to
       pretty_print: if True, will use nicer formatting
    '''
    with open(filename, mode) as filey:
        if print_pretty:
            filey.writelines(print_json(json_obj))
        else:
            filey.writelines(json.dumps(json_obj))
    return filename


def print_json(json_obj):
    ''' just dump the json in a "pretty print" format
    '''
    return json.dumps(
                    json_obj,
                    indent=4,
                    separators=(
                        ',',
                        ': '))


def read_file(filename, mode="r", readlines=True):
    '''write_file will open a file, "filename" and write content, "content"
       and properly close the file
    '''
    with open(filename, mode) as filey:
        if readlines is True:
            content = filey.readlines()
        else:
            content = filey.read()
    return content


def read_json(filename, mode='r'):
    '''read_json reads in a json file and returns
       the data structure as dict.
    '''
    with open(filename, mode) as filey:
        data = json.load(filey)
    return data



'''

Copyright (C) 2018-2019 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The SIF header format is developed and licensed by Sylabs,
see https://github.com/sylabs/sif/blob/master/pkg/sif/sif.go

'''

from sif.logger import bot
from sif.defaults import SIF_VERSION

arches = {
    "00": "Undefined/Unsupported architecture",
    "01": "386 (i[3-6]86) arch code",
    "02": "AMD64 arch code",
    "03": "ARM arch code",
    "04": "AARCH64 arch code",
    "05": "PowerPC 64 arch code",
    "06": "PowerPC 64 little-endian arch code",
    "07": "MIPS arch code",
    "08": "MIPS little-endian arch code",
    "09": "MIPS64 arch code",
    "10": "MIPS64 little-endian arch code",
    "11": "IBM s390x arch code"
}

class HeaderBase:
    '''A SIF Header is a binary header to describe a SIF image, and is defined
       by a set of pre-defined sections and a "SIF_MAGIC" header. See
       https://github.com/sylabs/sif/blob/master/pkg/sif/sif.go for details
    '''
    EndChar           = '\0'
    HdrLaunch         = "#!/usr/bin/env run-singularity\n"
    HdrMagic          = "SIF_MAGIC"
    HdrVersion        = "02"        # SIF SPEC VERSION
    HdrArchUnknown    = "00"        # Undefined/Unsupported arch
    HdrArch386        = "01"        # 386 (i[3-6]86) arch code
    HdrArchAMD64      = "02"        # AMD64 arch code
    HdrArchARM        = "03"        # ARM arch code
    HdrArchARM64      = "04"        # AARCH64 arch code
    HdrArchPPC64      = "05"        # PowerPC 64 arch code
    HdrArchPPC64le    = "06"        # PowerPC 64 little-endian arch code
    HdrArchMIPS       = "07"        # MIPS arch code
    HdrArchMIPSle     = "08"        # MIPS little-endian arch code
    HdrArchMIPS64     = "09"        # MIPS64 arch code
    HdrArchMIPS64le   = "10"        # MIPS64 little-endian arch code
    HdrArchS390x      = "11"        # IBM s390x arch code
    HdrLaunchLen      = 32          # len("#!/usr/bin/env... ")
    HdrMagicLen       = 10          # len("SIF_MAGIC")
    HdrVersionLen     = 3           # len("99")
    HdrArchLen        = 3           # len("99")
    DescrNumEntries   = 48          # default total # available descriptors
    DescrGroupMask    = 0xf0000000         # groups start at that offset
    DescrUnusedGroup  = DescrGroupMask     # descriptor without a group
    DescrDefaultGroup = DescrGroupMask | 1 # first groupid number created
    DescrUnusedLink   = 0                  # descriptor without link to other
    DescrEntityLen    = 256                # len("Joe Bloe <jbloe@gmail.com>...")
    DescrNameLen      = 128                # descriptor name (string identifier)
    DescrMaxPrivLen   = 384                # size reserved for descriptor specific data
    DescrStartOffset  = 4096               # descriptors start after global header
    DataStartOffset   = 32768              # data object start after descriptors

    def __init__(self, updates={}):
        '''a HeaderBase is a pre-set collection of locations and values
            for SIF headers. We use this to parse the header in python
        '''
        for key, val in updates.items():
            self.set_attribute(key, key)    

    def __str__(self):
        return "SIFHeader Version %s" % self.HdrVersion

    def __repr__(self):
        return self.__str__()

    def get_attribute(self, key, default):
        '''get an attribute from the base, or return a default if not defined.

           Parameters
           ==========
           key: the attribute name to get
           default: the default to return
        '''
        if hasattr(self, key):
            return getattr(self, key)
        return default

    def set_attribute(self, key, value):
        '''set an attribute for the base, only if it's already existing.

           Parameters
           ==========
           key: the attribute name to get
           value: the value to set
        '''
        if hasattr(self, key):
            setattr(self, key, value)



'''

Copyright (C) 2018-2019 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The SIF header format is developed and licensed by Sylabs,
see https://github.com/sylabs/sif/blob/master/pkg/sif/sif.go

'''

class Descriptor:
    '''A SIF Descriptor is the base, from which we derive 
       a Deffile, Signature, and Partition block. We read the descriptors after 
       the global header See
       https://github.com/sylabs/sif/blob/master/pkg/sif/sif.go for details
       This isn't a class with functions, just an organizational structure
       for fields and the format string.    

       # Datatype  int32   <1i
       # Used      bool    ?
       # ID        uint32  I 
       # Groupid   uint32  I 
       # Link      uint32  I 
       # Fileoff   int64   q
       # Filelen   int64   q
       # Storelen  int64   q
       # Ctime     int64   q       
       # Mtime     int64   q              
       # UID       int64   q              
       # Gid       int64   q              

       # TODO these two aren't parsed yet
       # Name  [DescrNameLen]byte    self.base.DescrNameLen
       # Extra [DescrMaxPrivLen]byte // big enough for extra data below

    '''

    name = 'Descriptor'
    fields = [ "Datatype",  
               "Used",
               "ID",
               "Groupid",
               "Link",
               "Fileoff",
               "Filelen",
               "Storelen",
               "Ctime",
               "Mtime",
               "UID",
               "Gid"     ]

    # Format string to read in above
    fmt = '<i?3I7q'

    def __str__(self):
        return "SIF Descriptor version 02 %s" % self.name

    def __repr__(self):
        return self.__str__()


class Deffile(Descriptor):
    '''A SIF Deffile is the first descriptor. It is essentially a Descriptor.
    '''
    name = 'Deffile'

    def __init__(self):
        Descriptor.__init__(self)


class Partition(Descriptor):
    '''A SIF Partition is the third block (id 2 at index 1). It has,
       in addition to the same fields, a fstype, parttype, and content.
       I'm not sure what content is, but the fstype and partype (I think)
       are under "extra"

           fstype: Squashfs   int32
           parttype: System   int32
           content: Linux     (not sure)
    '''
    name = 'Partition'

    def __init__(self):
        Descriptor.__init__(self)


class Signature(Descriptor):
    '''A SIF Signature is the fourth block (third descriptor) 
       It has the following fields under "extra"

           hashtype: SHA384
           entity: @ 
    '''
    name = 'Signature'

    def __init__(self):
        Descriptor.__init__(self)



'''

Copyright (C) 2018 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The SIF header format is developed and licensed by Sylabs,
see https://github.com/sylabs/sif/blob/master/pkg/sif/sif.go

'''

from sif.logger import bot
from sif.defaults import SIF_VERSION

# We don't have multiple versions, but can add loading logic here, e.g.
# We also might want (or need) to reload an image based on finding
# a different version.

def get_structure(version=None):
    '''get_structure will return known organization based on a version.
       The versions are organized into the subdirectories here. If the user
       doens't provide a version, we default to SIF_VERSION

       Parameters
       ==========
       version: a version to load, should be a corresponding string (e.g., 02)
    '''

    # Return a simple data structure with what we need
    class SIF:
        HeaderBase = None
        arches = None
        Deffile = None

    # If no provided version, use default
    if version == None:
        version = SIF_VERSION

    if version == "02":
        from .v02 import ( HeaderBase, 
                           arches, 
                           Deffile, 
                           Partition,
                           Signature )

    SIF.HeaderBase = HeaderBase()
    SIF.arches = arches
    SIF.Deffile = Deffile()
    SIF.Partition = Partition()
    SIF.Signature = Signature()
    return SIF



    def load_next(self, until_char=None):
        '''read in the header, byte by byte, and go until we hit the 
           null terminating character \0
        '''
        # If not defined, load until terminating string
        if until_char is None:
            until_char = self.base.EndChar

        byte = filey.read(1)
        item = b''

        # We will use the number to decode it
        number = 1
        while byte:
            item = item + byte
            byte = filey.read(1)
            number += 1
            if byte == until_char:
                break

        # The length is the number of characters we've read
        fmt = '%ss' % number
        return self.unpack(fmt, item)


    def _load_signatures(self, filey=None, close_file=False):

        # By default, we assume that we will not close the file handle
        close_file = False
 
        # Unless it's not provided, we need to clean up
        if filey == None:
            filey = open(self.image, 'rb')
            close_file = True

        offset = ( self.meta['descroff'] +      # start of offset
                   self.base.DescrNameLen +     # length of name (128)
                   self.base.DescrMaxPrivLen )  # max private length

        # Let's try reading in different ones...
        for o in range(0, self.meta['descrlen']):
            offset = self.meta['descroff'] + o
            filey.seek(offset)
            values = self.unpack_bytes(filey, self.Deffile.fmt)
            if values[3] = [1, 2, 3]:
                print('offset %s, %s' %(offset, values))


        # let's try to find based on iota increments
        for o in range(0, self.meta['descrlen'] * 10):
            offset = self.meta['descroff'] + o
            filey.seek(offset)
            values = self.unpack_bytes(filey, part.fmt)
            if int(values[0]/1000) == 16:
                print('offset %s, %s' %(offset, values))



        filey.seek(self.Signature.start)

        # Brute force it again 
        for o in range(0, self.meta['datalen']):
            offset = self.Signature.start + o
            filey.seek(offset)
            values = self.unpack_bytes(filey, self.Signature.fmt)
            if values[0] != 0:
                try:
                    when = datetime.datetime.fromtimestamp(values[8])
                    if when.year > 2015 and when.year < 2019:
                        print('Offset %s: %s' % (offset, datetime.datetime.fromtimestamp(values[8])))
                    #print('offset %s, %s' %(offset, values))
                except:
                    pass

        # see sif.header module for the default fields and format strings
        values = self.unpack_bytes(filey, self.Signature.fmt)
         
        # Update the descriptors dictionary
        for d in range(len(values)):
            partition[self.Partition.fields[d]] = values[d]

        # squashfs-955608129.img
        name = self.read_and_strip(filey, "%sc" % self.base.DescrNameLen)
        partition['name'] = name

        # I think partype and fstype might be part of extra?
        fstype, partype = self.unpack_bytes(filey, "2i")
        partition['fstype'] = fstype        
        partition['partype'] = partype

        # The remaining extra is the self.base.DescrMaxPrivLen - len(2i)
        fmt = '%sc' % (self.base.DescrMaxPrivLen - calcsize('2i'))
       
        # Can we get a name (this seems wrong) !
        extra = self.read_and_strip(filey, fmt)

        partition['extra'] = extra

        # Can we find the start of the signature?
        self.Signature.start = filey.tell()




'''

Copyright (C) 2018-2019 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from sif.logger import bot
from sif.utils import ( 
    get_userhome, 
    read_json, 
    getenv,
    convert2boolean 
)
import tempfile
import os
import sys


#########################
# Caches
#########################

USERHOME = get_userhome()
SIF_VERSION = getenv("SIF_VERSION", "02")

# Singularity Cache
_cache = os.path.join(USERHOME, ".singularity")
SINGULARITY_CACHE = getenv("SINGULARITY_CACHEDIR", default=_cache)

# Temporary Storage
SIF_TMPDIR = os.environ.get('SIF_TMPDIR', tempfile.gettempdir())




# Copyright (C) 2018-2019 Vanessa Sochat.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.



__version__ = "0.0.11"
AUTHOR = 'Vanessa Sochat'
AUTHOR_EMAIL = 'vsochat@stanford.edu'
NAME = 'sif'
PACKAGE_URL = "http://www.github.com/singularityhub/sif"
KEYWORDS = 'container, singularity, sif'
DESCRIPTION = "python handle for reading sif images"
LICENSE = "LICENSE"

INSTALL_REQUIRES = (
)



'''

Copyright (C) 2018-2019 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from sif.logger import bot
import hashlib
import os
import re


def get_image_hash(image_path):
    '''return an md5 hash of the file based on a criteria level. This
    is intended to give the file a reasonable version.
    :param image_path: full path to the singularity image
    '''
    hasher = hashlib.md5()
    with open(image_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hasher.update(chunk)
    return hasher.hexdigest()



def parse_image_name(image_name,
                     tag=None,
                     version=None, 
                     defaults=True, 
                     ext="simg",
                     default_collection="library",
                     default_tag="latest",
                     base=None):

    '''return a collection and repo name and tag
    for an image file.
    
    Parameters
    =========
    image_name: a user provided string indicating a collection,
                image, and optionally a tag.
    tag: optionally specify tag as its own argument
         over-rides parsed image tag
    defaults: use defaults "latest" for tag and "library"
              for collection. 
    base: if defined, remove from image_name, appropriate if the
          user gave a registry url base that isn't part of namespace.

    '''
    if base is not None:
        image_name = image_name.replace(base,'').strip('/')

    result = dict()
    image_name = re.sub('[.](img|simg)','',image_name)
    image_name = re.split('/', image_name, 1)

    # User only provided an image
    if len(image_name) == 1:
        collection = ''
        if defaults is True:
            collection = default_collection
        image_name = image_name[0]

    # Collection and image provided
    elif len(image_name) >= 2:
        collection = image_name[0].lower()
        image_name = image_name[1]
    
    # Is there a version?
    image_name = image_name.split('@')
    if len(image_name) > 1: 
        version = image_name[1].lower()
    image_name = image_name[0]

    # Is there a tag?
    image_name = image_name.split(':')

    # No tag in provided string
    if len(image_name) > 1: 
        tag = image_name[1]
    image_name = image_name[0].lower()
    
    # If still no tag, use default or blank
    if tag is None and defaults is True:
        tag = default_tag

    # Piece together the filename
    uri = "%s/%s" % (collection, image_name)    
    url = uri
    if tag is not None:
        uri = "%s-%s" % (uri, tag)
    if version is not None:
        uri = "%s@%s" % (uri, version)

    storage = "%s.%s" %(uri, ext)
    result = {"collection": collection,
              "image": image_name,
              "url": url,
              "tag": tag,
              "version": version,
              "storage": storage,
              "uri": uri}

    return result

def format_container_name(name, special_characters=None):
    '''format_container_name will take a name supplied by the user,
    remove all special characters (except for those defined by "special-characters"
    and return the new image name.
    '''
    if special_characters is None:
        special_characters = []
    return ''.join(e.lower()
                   for e in name if e.isalnum() or e in special_characters)


def get_uri(image):
    '''get the uri for an image, if within acceptable
 
       Parameters
       ==========
       image: the image uri, in the format <uri>://<registry>/<namespace>:<tag>

    '''
    # Ensure we have a string
    image = image or ''

    # Find uri prefix, including ://
    regexp = re.compile('^.+://')
    uri = regexp.match(image)

    if uri is not None:
        uri = (uri.group().lower()
                          .replace('_','-')
                          .replace('://',''))
 
        accepted_uris = ['aws',
                         'ceph',
                         'docker', 
                         'dropbox',
                         'hub',
                         'gitlab',
                         'globus',
                         'registry', 
                         'nvidia', 
                         'google-storage',
                         'google-drive']

        # Allow for Singularity compatability
        if uri == "shub": uri = "hub"

        if uri not in accepted_uris:
            bot.warning('%s is not a recognized uri.' %uri)
            uri = None

    return uri


def remove_uri(image):
    '''remove_uri will remove the uri from the image path, if provided.
 
       Parameters
       ==========
       image: the image uri, in the format <uri>://<registry>/<namespace>:<tag>

    '''
    return re.sub('^.+://','', image)




# Copyright (C) 2018-2019 Vanessa Sochat.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import uuid
from sif.logger import bot
from sif.defaults import SIF_VERSION
from struct import ( unpack, calcsize )

class SIFHeader:

    def __init__(self, image, load_header=True, version=None):

        # Load the base for a particular SIF version
        self.load_base(version)

        # We will store global headers, and descriptors
        self.meta = dict()
        self.desc = dict()

        if not os.path.exists(image):
            bot.exit('Cannot find %s.' % image)

        if not self.is_sif(image):
            bot.exit('%s is not a SIF file.' % image)

        bot.info('%s is a SIF file.' % image)
        self.image = image

        # The user might want to wait to load
        if load_header is True:
            self.load_header()


    def __str__(self):
        '''string representation of a SIF image shows the path'''
        return "<SIF:%s>" % self.image

    def __repr__(self):
        '''representation of SIF image is also the path'''
        return self.__str__()


################################################################################
# Printing
################################################################################


    def print_header(self):
        '''print the metadata for the user to see
        '''
        for key, val in self.meta.items():
            bot.info('SIF Header %s %s' % (key, val) )
        bot.newline()

    def print_descriptor(self, descriptor, skip_keys=None):
        '''print the definition file metadata for the user to see

           Parameters
           ==========
           name: the key in self.desc to iterate over
           skip_keys: skip these keys in self.desc[key]
        ''' 
        if skip_keys == None:
            skip_keys = []

        if not isinstance(skip_keys, list):
            skip_keys = [skip_keys]

        if descriptor in self.desc:
            name = descriptor.capitalize()
            for key, val in self.desc[descriptor].items():
                if key not in skip_keys:
                    bot.info('%s %s %s' % (name, key, val) )
        bot.newline()


    def print_descriptor_deffile(self):
        self.print_descriptor('deffile', 'content')

    def print_descriptor_partition(self):
        self.print_descriptor('partition')

    def print_descriptor_signature(self):
        self.print_descriptor('signature')

    def print_deffile(self):
        '''print the definition file for the user to see
        '''
        if 'deffile' in self.desc:
            if 'content' in self.desc['deffile']:
                print(self.desc['deffile']['content'])

    def print_arch(self):
        '''print the human friendly architecture'''
        if 'arch' in self.meta:
            if self.meta['arch'] in self.arches:
                bot.info('Architecture: %s' % self.arches[self.meta['arch']])
            bot.newline()

################################################################################
# Validation
################################################################################


    def is_sif(self, image):
        '''determine if an image is SIF based on finding SIF_MAGIC
        ''' 

        with open(image, 'rb') as filey:

            # Skip the first line
            filey.seek(self.base.HdrLaunchLen)

            # Read # of bytes corresponding to length of sif magic
            fmt = "%ss" % self.base.HdrMagicLen
 
            # Read length of header magic (if fails, not sif)           
            try:
                line = self.unpack_chars(filey, fmt)
            except UnicodeDecodeError:
                pass
                return False

            if line != None:
                if isinstance(line, str) and line.startswith(self.base.HdrMagic):
                    return True

            return False


    def unpack_chars(self, filey, fmt, number=None, encoding='utf-8'):
        '''unpack characters, and handle the encoding and removing
           an end character.
        '''
        line, = self.unpack_bytes(filey, fmt, number)

        # Return numbers, otherwise parse bytes to string
        if isinstance(line, (int, float)):
            return line

        return line.decode(encoding).replace(self.base.EndChar, '')


################################################################################
# Loading
################################################################################

    def load_base(self, version=None):
        '''load a sif base, or default to version 02
        '''
        from sif.header import get_structure
        SIF = get_structure(version)

        # We have a header base, arches
        self.base = SIF.HeaderBase
        self.arches = SIF.arches

        # Descriptors
        self.Deffile = SIF.Deffile 
        self.Partition = SIF.Partition
        self.Signature = SIF.Signature

    def read_and_strip(self, filey, fmt, number=None):
        '''read a number of bytes (number of based on a format) from the
           file, remove empty / white spaces indicated by \0.
        '''
        values = self.read_bytes(filey, fmt, number)
         
        try:
            return values.decode('utf-8').replace('\0', '')
        except:
            return values


    def read_bytes(self, filey, fmt, number=None):
        '''just read a number of bytes from an open file.

           Parameters
           ==========
           filey: an open file object
           number: the number of bytes to read
           fmt: an optional format string
        '''
        if number == None:
            number = calcsize(fmt)

        return filey.read(number)


    def unpack_bytes(self, filey, fmt, number=None):
        '''read AND unpack a number of bytes from an open file. If a 
           format is provided, convert to utf-8 and return the string. 
           If fmt is None, assume a character string of the same length.

           Parameters
           ==========
           filey: an open file object
           number: the number of bytes to read
           fmt: an optional format string
        '''
        byte_values = self.read_bytes(filey, fmt, number)
        return unpack(fmt, byte_values)



################################################################################
# Global Header
################################################################################

            
    def load_header(self):
        '''Once we know is sif, load the header.
        ''' 

        # Start at end of interpreter line, and after sif magix
        byte_start = self.base.HdrLaunchLen + self.base.HdrMagicLen

        with open(self.image, 'rb') as filey:

            # Skip the first line
            filey.seek(byte_start)

            # 1. Read the version
            self.meta['version'] = self.unpack_chars(filey, 
                                                     number=self.base.HdrVersionLen, 
                                                     fmt = "%ss" % self.base.HdrVersionLen)

            # 2. Read the architecture
            self.meta['arch'] = self.unpack_chars(filey, 
                                                  number=self.base.HdrArchLen,
                                                  fmt = "<%ss" % self.base.HdrArchLen)

            # 3. Read the uuid - this returns little endian bytes
            uuid_bytes = self.read_bytes(filey, fmt='<16c')

            # Let the uuid library read the binary data for us!
            self.meta['uuid'] = str(uuid.UUID(bytes_le=uuid_bytes))

            # 4. Read in the remaining Global header fields, 1 signed int
            items = ['ctime', 
                     'mtime', 
                     'dfree',
                     'dtotal',
                     'descroff',
                     'descrlen', 
                     'dataoff', 
                     'datalen']

            for item in items:
                self.meta[item] = self.unpack_bytes(filey, fmt='1q')[0]

            # Load the definition file descriptors
            self.desc['deffile'] = self._load_deffile()
            self.desc['partition'] = self._load_partition()
            self.desc['signature'] = self._load_signature()

        # Update the user with what was loaded
        self.print_header()
        self.print_arch()
        self.print_descriptor_deffile() 
        self.print_descriptor_partition()
        self.print_descriptor_signature()

################################################################################
# Descriptors
################################################################################


    def _load_deffile(self):
        ''' load the header descriptor for the definition file. If a file
            object is not provided, open. The fields and format string are 
            provided via the self.Deffile object, which we get from 
            self.header --> __init__.py --> get_structure() --> SIF
        '''

        filey = open(self.image, 'rb')

        # Read to the description offset
        descriptors = dict()
        filey.seek(self.meta['descroff'])

        # see sif.header module for the default fields and format strings
        values = self.unpack_bytes(filey, self.Deffile.fmt)
         
        # Update the descriptors dictionary
        for d in range(len(values)):
            descriptors[self.Deffile.fields[d]] = values[d]

        # Can we get a name (this seems wrong) - loads as "."
        name = self.read_and_strip(filey, "%sc" % self.base.DescrNameLen)
        extra = self.read_and_strip(filey, '%sc' % self.base.DescrMaxPrivLen)

        descriptors['name'] = name
        descriptors['extra'] = extra

        # Save the location for the start of the partition
        self.Partition.start = filey.tell()

        # Definition File - offset should point us to the deffile    
        filey.seek(descriptors['Fileoff'])

        # And the length is also provided
        fmt = '%sc' % descriptors['Filelen']
        number = calcsize(fmt)
        deffile = filey.read(number)

        # Try to decode to utf-8 for the user
        try:
            deffile = deffile.decode('utf-8')
        except:
            pass

        descriptors['content'] = deffile
        filey.close()

        return descriptors


    def _load_partition(self):
        ''' load the paritition descriptor. We must get the start based on
            first reading the deffile desciptor, which adds the location
            to self.Parition.start. If we don't have this attribute, we need
            to run it first (This should not happen in normal cases, as they
            are called under one function, one after the other)
        '''
        if not hasattr(self.Partition, 'start'):
            self._load_deffile()

        filey = open(self.image, 'rb')

        partition = dict()
        filey.seek(self.Partition.start)

        # see sif.header module for the default fields and format strings
        values = self.unpack_bytes(filey, self.Deffile.fmt)
         
        # Update the descriptors dictionary
        for d in range(len(values)):
            partition[self.Partition.fields[d]] = values[d]

        # squashfs-955608129.img
        name = self.read_and_strip(filey, "%sc" % self.base.DescrNameLen)
        partition['name'] = name

        # partype and fstype might be part of extra?
        fstype, partype = self.unpack_bytes(filey, "2i")
        partition['fstype'] = fstype        
        partition['partype'] = partype

        # The remaining extra is the self.base.DescrMaxPrivLen - len(2i)
        fmt = '%sc' % (self.base.DescrMaxPrivLen - calcsize('2i'))       
        extra = self.read_and_strip(filey, fmt)

        partition['extra'] = extra

        # Can we find the start of the signature?
        self.Signature.start = filey.tell()
        filey.close()
        return partition


    def _load_signature(self):
        '''finally, load the signature descriptor. We again get the start
           based on loading the partition first.
        '''
        if not hasattr(self.Signature, 'start'):
            self._load_partition()

        filey = open(self.image, 'rb')

        signature = dict()
        filey.seek(self.Signature.start)
        values = self.unpack_bytes(filey, self.Signature.fmt)

        # If the first value is 0, the container isn't signed
        if values[0] != 0:

            for d in range(len(values)):
                signature[self.Signature.fields[d]] = values[d]

            # squashfs-955608129.img
            name = self.read_and_strip(filey, "%sc" % self.base.DescrNameLen)
            signature['name'] = name
            hashtype, = self.unpack_bytes(filey, "i") # unpack tuple

            # We need to parse the entity here (I don't know what that means)
            # see self.base.DescEntityLen

            # Go to the signature block, and retrieve it
            filey.seek(signature['Fileoff'])
            fmt = '%sc' % signature['Filelen']
            signed = self.read_and_strip(filey, fmt)

            signature['hastype'] = hashtype
            signature['publicKey'] = signed

        filey.close()
        return signature



'''

Copyright (C) 2018-2019 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''

from sif.logger import bot
from .fileio import ( mkdir_p, get_userhome )
import tempfile
import os
import pwd
import sys


def clean_path(path):
    '''clean_path will canonicalize the path

       Parameters
       ==========
       path: the path to clean
    '''
    return os.path.realpath(path.strip(" "))


def convert2boolean(arg):
    '''convert2boolean is used for environmental variables that must be
       returned as boolean

       Parameters
       ==========
       arg: the string argument to convert to boolean
    '''
    if not isinstance(arg, bool):
        return arg.lower() in ("yes", "true", "t", "1", "y")
    return arg


def getenv(variable_key, default=None, required=False, silent=False):
    '''getenv will attempt to get an environment variable. If the variable
        is not found, None is returned.

        Parameters
        ==========
        variable_key: the variable name
        required: exit with error if not found
        silent: Do not print debugging information for variable
    '''
    variable = os.environ.get(variable_key, default)
    if variable is None and required:
        bot.error(
            "Cannot find environment variable %s, exiting." %
            variable_key)
        sys.exit(1)

    if silent:
        bot.verbose2("%s found" % variable_key)
    else:
        if variable is not None:
            bot.verbose2("%s found as %s" % (variable_key, variable))
        else:
            bot.verbose2("%s not defined (None)" % variable_key)

    return variable


def get_cache(subfolder=None, quiet=False):
    '''get_cache will return the user's cache for singularity.

       Parameters
       ==========
       subfolder: a subfolder in the cache base to retrieve, specifically
       quiet: an extra controller for verbositry (suppress output)
    '''

    DISABLE_CACHE = convert2boolean(getenv("SINGULARITY_DISABLE_CACHE",
                                           default=False))
    if DISABLE_CACHE:
        SINGULARITY_CACHE = tempfile.mkdtemp()
    else:
        userhome = pwd.getpwuid(os.getuid())[5]
        _cache = os.path.join(userhome, ".singularity")
        SINGULARITY_CACHE = getenv("SINGULARITY_CACHEDIR", default=_cache)

    # Clean up the path and create
    cache_base = clean_path(SINGULARITY_CACHE)

    # Does the user want to get a subfolder in cache base?
    if subfolder is not None:
        cache_base = "%s/%s" % (cache_base, subfolder)

    # Create the cache folder(s), if don't exist
    mkdir_p(cache_base)

    if not quiet:
        bot.debug("Cache folder set to %s" % cache_base)
    return cache_base



#!/usr/bin/env python

# Copyright (C) 2019 Vanessa Sochat.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from sif.logger import bot
import sys
import os

def main(args):

    lookup = { 'ipython': ipython,
               'python': python,
               'bpython': bpython }

    shells = ['ipython', 'python', 'bpython']

    # The user must supply an image
    if args.image is None:
        bot.exit('You must provide an image to load a SIFHeader client.')

    image = args.image[0]

    # The image must exist
    if not os.path.exists(image):
        bot.exit('Cannot find %s' % image)

    # Present order of liklihood to have on system
    for shell in shells:
        try:
            return lookup[shell](image)
        except ImportError:
            pass
    

def ipython(image):
    '''give the user an ipython shell, optionally load image
    '''
    from sif.main import SIFHeader  
    header = SIFHeader(image)  
    from IPython import embed
    embed()


def bpython(image):
    import bpython
    from sif.main import SIFHeader  
    header = SIFHeader(image)  
    bpython.embed(locals_={'header': header,
                           'image': image,
                           'SIFHeader': SIFHeader})

def python(image):
    import code
    from sif.main import SIFHeader  
    header = SIFHeader(image)  
    code.interact(local={"header": header,
                         'image': image,
                         'SIFHeader': SIFHeader})



'''

Copyright (C) 2018-2019 Vanessa Sochat.

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''


from setuptools import setup, find_packages
import codecs
import os

##########################################################################################
# HELPER FUNCTIONS #######################################################################
##########################################################################################

def get_lookup():
    '''get version by way of singularity.version, returns a 
    lookup dictionary with several global variables without
    needing to import singularity
    '''
    lookup = dict()
    version_file = os.path.join('sif', 'version.py')
    with open(version_file) as filey:
        exec(filey.read(), lookup)
    return lookup

# Read in requirements
def get_requirements(lookup=None):
    '''get_requirements reads in requirements and versions from
    the lookup obtained with get_lookup'''

    if lookup == None:
        lookup = get_lookup()

    install_requires = []
    for module in lookup['INSTALL_REQUIRES']:
        module_name = module[0]
        module_meta = module[1]
        if "exact_version" in module_meta:
            dependency = "%s==%s" %(module_name,module_meta['exact_version'])
        elif "min_version" in module_meta:
            if module_meta['min_version'] == None:
                dependency = module_name
            else:
                dependency = "%s>=%s" %(module_name,module_meta['min_version'])
        install_requires.append(dependency)
    return install_requires



# Make sure everything is relative to setup.py
install_path = os.path.dirname(os.path.abspath(__file__)) 
os.chdir(install_path)

# Get version information from the lookup
lookup = get_lookup()
VERSION = lookup['__version__']
NAME = lookup['NAME']
AUTHOR = lookup['AUTHOR']
AUTHOR_EMAIL = lookup['AUTHOR_EMAIL']
PACKAGE_URL = lookup['PACKAGE_URL']
KEYWORDS = lookup['KEYWORDS']
DESCRIPTION = lookup['DESCRIPTION']
LICENSE = lookup['LICENSE']
with open('README.md') as filey:
    LONG_DESCRIPTION = filey.read()

##########################################################################################
# MAIN ###################################################################################
##########################################################################################


if __name__ == "__main__":

    INSTALL_REQUIRES = get_requirements(lookup)

    setup(name=NAME,
          version=VERSION,
          author=AUTHOR,
          author_email=AUTHOR_EMAIL,
          maintainer=AUTHOR,
          maintainer_email=AUTHOR_EMAIL,
          packages=find_packages(), 
          include_package_data=True,
          zip_safe=False,
          url=PACKAGE_URL,
          license=LICENSE,
          description=DESCRIPTION,
          long_description=LONG_DESCRIPTION,
          keywords=KEYWORDS,
          setup_requires=["pytest-runner"],
          tests_require=["pytest"],
          install_requires = INSTALL_REQUIRES,
          classifiers=[
              'Intended Audience :: Science/Research',
              'Intended Audience :: Developers',
              'Programming Language :: Python',
              'Topic :: Software Development',
              'Topic :: Scientific/Engineering',
              'Operating System :: Unix',
              'Programming Language :: Python :: 2.7',
              'Programming Language :: Python :: 3',
          ],

          entry_points = {'console_scripts': [ 'sif=sif.client:main' ] })



from .terminal import (
    run_command,
    check_install,
    get_installdir,
    which
)
from .cache import ( get_cache, getenv, convert2boolean )
from .fileio import (
    get_userhome,
    get_tmpdir,
    get_tmpfile,
    mkdir_p,
    print_json,
    read_file,
    read_json,
    write_file,
    write_json
)
